"use strict";
/**
 * # promisemitter
 *
 * An observer / event emitter implementation with promise support.
 *
 * Its meant to behave similarly to typical synchronous
 * [Java / NET observers](https://msdn.microsoft.com/en-us/library/ff648108.aspx),
 * where once you notify your subscribers you are able to wait for their update
 * functions to execute before proceeding
 *

# Example

Given a blog post creator:

```typescript
import po = require('promise-observer')

class BlogPostCreator {
    postCreated = po.create();

    create(blogPost) {
      actuallyCreateBlogpost()
        .then(post => this.postCreated.emit(post))
        // wait for all attached events to complete before commiting.
        .then(commitTransaction);
    }
}

```

a categorizer can attach to its events

```typescript
postCategorized = blogPostCreator.postCreated.subscribe(post =>
  categorize(post).then(saveCategory).thenReturn(post));
```

an indexer can add search terms to the index for that post

```typescript
postIndexed = blogPostCreator.postCreated.subscribe(post =>
  index(post).then(saveIndex).thenReturn(post));
```

Then, the email notification system can wait for the post to be
categorized and indexed before sending a notification to all subscribers:

```typescript
onPostNotification = blogPostCreator.postCreated.subscribe(post => {
  var categorized = postCategorized.next(categorizedPost => categorizedPost.id == post.id);
  var indexed = postIndexed.next(indexedPost => indexedPost.id == post.id);
  return Promise.join(categorized, indexed, _ => sendEmailNotification(post))
});
```

 */
var __extends = (this && this.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var helpers = require("./helpers");
var TimeoutError = /** @class */ (function (_super) {
    __extends(TimeoutError, _super);
    function TimeoutError() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        /**
         * Stringified code of the listener that threw the error
         */
        _this.timedoutListener = '<unknown>';
        return _this;
    }
    return TimeoutError;
}(Error));
exports.TimeoutError = TimeoutError;
function timeout(p, delay) {
    var e = new TimeoutError('Timeout waiting for event listener to complete');
    var timeout = new Promise(function (_, rej) {
        return setTimeout(function () {
            rej(e);
        }, delay);
    });
    return Promise.race([timeout, p]);
}
/**
 *
 * Creates a new emitter. The emitter contains emit and subscribe methods. The subscribe method
 * is also an `Observable<T>`
 * @param opts - Creation options
 */
function create(opts) {
    opts = opts || {};
    var subscriptions = [];
    function emit(t) {
        opts = opts || {};
        var results = [];
        for (var k = 0; k < subscriptions.length; k) {
            var current = subscriptions[k];
            var emitPromise = Promise.resolve(current.emit(t));
            if (opts.emitTimeout != null)
                emitPromise = timeout(emitPromise, opts.emitTimeout).catch(function (e) {
                    if (e.timedoutListener === '<unknown>')
                        e.timedoutListener = String(current.listener);
                    throw e;
                });
            results.push(emitPromise);
            if (current === subscriptions[k])
                ++k;
        }
        return helpers.waitAll(results);
    }
    function next(predicate) {
        return new Promise(function (resolve) {
            var sub = subscribe(function (item) {
                if (predicate == null || predicate(item)) {
                    remove(sub);
                    resolve(item);
                }
            });
        });
    }
    function subscribe(listener) {
        var emitter = create();
        var sub = emitter.subscribe;
        subscriptions.push({
            emit: function (item) { return Promise.resolve(listener(item)).then(emitter.emit); },
            target: sub,
            listener: listener
        });
        sub.unlink = function () { return remove(sub); };
        return sub;
    }
    function remove(target) {
        for (var k = 0; k < subscriptions.length; ++k)
            if (subscriptions[k].target == target)
                return subscriptions.splice(k, 1);
    }
    return { emit: emit, subscribe: Object.assign(subscribe, { remove: remove, next: next }) };
}
exports.create = create;
//# sourceMappingURL=index.js.map